;; Generated package description from erc.el  -*- no-byte-compile: t -*-
(define-package "erc" "5.5.0.20230303.134918" "An Emacs Internet Relay Chat client" '((emacs "27.1") (compat "29.1.3.4")) :commit "49e021e5be943477329b065fbff76862ea832acc" :authors '(("Alexander L. Belikoff" . "alexander@belikoff.net")) :maintainer '("Amin Bandali <bandali@gnu.org>, F. Jason Park" . "jp@neverwas.me") :keywords '("irc" "chat" "client" "internet") :url "https://www.gnu.org/software/emacs/erc.html")
